@@ -1,5 +1,6 @@
 /*
  * Copyright 2012 Canonical Ltd.
+ * Copyright 2018 Sahil Arora <sahilarora.535@gmail.com>
  *
  * This program is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 3, as published
@@ -23,7 +24,7 @@
 extern "C" {
 #endif
 
-typedef struct PDFDoc pdf_t;
+typedef struct QPDF pdf_t;
 
 typedef struct _opt opt_t;
 
@@ -39,10 +40,10 @@ struct _opt {
 pdf_t * pdf_load_template(const char *filename);
 void pdf_free(pdf_t *pdf);
 void pdf_write(pdf_t *doc, FILE *file);
-void pdf_prepend_stream(pdf_t *doc, int page, char *buf, size_t len);
-void pdf_add_type1_font(pdf_t *doc, int page, const char *name);
-void pdf_resize_page (pdf_t *doc, int page, float width, float length, float *scale);
-void pdf_duplicate_page (pdf_t *doc, int page, int count);
+void pdf_prepend_stream(pdf_t *doc, unsigned page, char const *buf, size_t len);
+void pdf_add_type1_font(pdf_t *doc, unsigned page, const char *name);
+void pdf_resize_page(pdf_t *doc, unsigned page, float width, float length, float *scale);
+void pdf_duplicate_page (pdf_t *doc, unsigned page, unsigned count);
 int pdf_fill_form(pdf_t *doc, opt_t *opt);
 
 #ifdef __cplusplus
@@ -50,4 +51,3 @@ int pdf_fill_form(pdf_t *doc, opt_t *opt);
 #endif
 
 #endif
-
